@@ -35,6 +35,14 @@ static void gasnetc_atexit(void);
 
 gasneti_handler_fn_t gasnetc_handler[GASNETC_MAX_NUMHANDLERS]; /* handler table (recommended impl) */
 
+#if GASNETC_GNI_FIREHOSE
+static int gasnetc_did_firehose_init = 0;
+static firehose_info_t gasnetc_firehose_info;
+size_t gasnetc_fh_align;
+size_t gasnetc_fh_align_mask;
+int gasnetc_use_firehose = 1;
+#endif
+
 /* ------------------------------------------------------------------------------------ */
 /*
   Initialization
@@ -108,6 +116,32 @@ static int gasnetc_bootstrapInit(int *argc, char ***argv) {
   gasnetc_ptag    = (uint8_t)  atoi(getenv("PMI_GNI_PTAG"));
   gasnetc_cookie  = (uint32_t) atoi(getenv("PMI_GNI_COOKIE"));
 
+  /* In GASNet+MPI hybrid applications in each lib must have a distinct cookie.
+   * There are three choice:
+   * 1) If MPI is always initialized first, then enable the first variant below
+   *    and (re)build the GASNet library.
+   * 2) If GASNet is always initialized first, then enable the second variant
+   *    below and (re)build the GASNet library.
+   * 3) If you don't want a build of GASNet that assumes the order the two
+   *    libaries will be initialized, then clone the third variant into the
+   *    application, to run between the two initializations.
+   */
+  #if 0   /* Option 1) If MPI is initialized first enable this: */
+    gasnetc_cookie += 1;
+  #endif
+  #if 0   /* Option 2) If GASNet is initialized first enable this: */
+  { char cookie[10];
+    sprintf(cookie, "%u", gasnetc_cookie + 1);
+    gasnett_setenv("PMI_GNI_COOKIE", cookie);
+  }
+  #endif
+  #if 0   /* Option 3) Copy this variant into your code between the initializations */
+  { char cookie[10];
+    sprintf(cookie, "%u", (1 + atoi(getenv("PMI_GNI_COOKIE"))));
+    gasnett_setenv("PMI_GNI_COOKIE", cookie);
+  }
+  #endif
+
   /* As good a place as any for this: */
   if (0 == gasneti_mynode) {
     static const char *old_vars[][3] = {
@@ -144,8 +178,6 @@ static int gasnetc_bootstrapInit(int *argc, char ***argv) {
  * Taken from vapi-conduit w/o the thread safetly complications 
  */
 
-static int gasnetc_bootstrap_am_coll = 0;
-
 static gasnet_node_t gasnetc_dissem_peers = 0;
 static gasnet_node_t *gasnetc_dissem_peer = NULL;
 static gasnet_node_t *gasnetc_exchange_rcvd = NULL;
@@ -157,10 +189,9 @@ static void gasnetc_sys_barrier_reqh(gasnet_token_t token, uint32_t arg)
     gasnetc_sys_barrier_rcvd[arg&1] |= arg;
 }
 
-GASNETI_NEVER_INLINE(gasnetc_bootstrapBarrier,
-void gasnetc_bootstrapBarrier(void))
+GASNETI_NEVER_INLINE(gasnetc_bootstrapBarrier_gni,
+void gasnetc_bootstrapBarrier_gni(void))
 {
-  if (gasnetc_bootstrap_am_coll) {
     static int phase = 0;
     int pre_attach = !gasneti_attach_done;
     int i;
@@ -195,9 +226,6 @@ void gasnetc_bootstrapBarrier(void))
     /* reset for next barrier */
     gasnetc_sys_barrier_rcvd[phase] = 0;
     phase ^= 1;
-  } else {
-    gasneti_bootstrapBarrier_pmi();
-  }
 }
 
 #define GASNETC_SYS_EXCHANGE_MAX GASNETC_GNI_MAX_MEDIUM
@@ -227,10 +255,9 @@ static void gasnetc_sys_exchange_reqh(gasnet_token_t token, void *buf,
   ++gasnetc_sys_exchange_rcvd[phase][step];
 }
 
-GASNETI_NEVER_INLINE(gasnetc_bootstrapExchange,
-void gasnetc_bootstrapExchange(void *src, size_t len, void *dest))
+GASNETI_NEVER_INLINE(gasnetc_bootstrapExchange_gni,
+void gasnetc_bootstrapExchange_gni(void *src, size_t len, void *dest))
 {
-  if (gasnetc_bootstrap_am_coll) {
     static int phase = 0;
 
     const int pre_attach = !gasneti_attach_done;
@@ -296,9 +323,6 @@ end_network_comms:
     gasneti_free(temp);
     gasnetc_sys_exchange_buf[phase] = NULL;
     phase ^= 1;
-  } else {
-    gasneti_bootstrapExchange_pmi(src, len, dest);
-  }
 
 #if GASNET_DEBUG
   /* verify own data as a sanity check */
@@ -387,12 +411,7 @@ done:
   gasnetc_handler[_hidx_gasnetc_sys_barrier_reqh]  = (gasneti_handler_fn_t)&gasnetc_sys_barrier_reqh;
   gasnetc_handler[_hidx_gasnetc_sys_exchange_reqh] = (gasneti_handler_fn_t)&gasnetc_sys_exchange_reqh;
 
-  gasnetc_bootstrap_am_coll = 1;
-#if GASNETC_USE_MULTI_DOMAIN
-  /* Need to use PMI exchange at each pthread spawn */
-#else
   gasneti_bootstrapCleanup_pmi(); /* No further use of PMI-based colelctives */
-#endif
 }
 
 static void gasnetc_sys_coll_fini(void)
@@ -512,8 +531,8 @@ extern uintptr_t gasnetc_MaxPinMem(uintptr_t msgspace)
   pm_limit -= msgspace;
 
   limit = gasneti_mmapLimit((uintptr_t)-1, pm_limit,
-                            &gasnetc_bootstrapExchange,
-                            &gasnetc_bootstrapBarrier);
+                            &gasnetc_bootstrapExchange_gni,
+                            &gasnetc_bootstrapBarrier_gni);
 
 
   if_pf (gasneti_getenv_yesno_withdefault("GASNET_PHYSMEM_NOPROBE", 0)) {
@@ -567,10 +586,6 @@ static int gasnetc_init(int *argc, char ***argv) {
   if (gasneti_init_done) 
     GASNETI_RETURN_ERRR(NOT_INIT, "GASNet already initialized");
 
-  gasneti_init_done = 1; /* enable early to allow tracing */
-
-  gasneti_freezeForDebugger();
-
 #if GASNET_DEBUG_VERBOSE
     /* note - can't call trace macros during gasnet_init because trace system not yet initialized */
     fprintf(stderr,"gasnetc_init(): about to call gasnetc_init...\n"); fflush(stderr);
@@ -579,47 +594,33 @@ static int gasnetc_init(int *argc, char ***argv) {
   ret = gasnetc_bootstrapInit(argc, argv);
   if (ret != GASNET_OK) return ret;
 
+  gasneti_init_done = 1; /* enable early to allow tracing */
+
+  gasneti_freezeForDebugger();
+
+  /* Now enable tracing of all the following steps */
+  gasneti_trace_init(argc, argv);
+
   #if GASNET_DEBUG_VERBOSE
     fprintf(stderr,"gasnetc_init(): gasnetc_init done - node %i/%i starting...\n", 
       gasneti_mynode, gasneti_nodes); fflush(stderr);
   #endif
 
-  /* determine which GASNet nodes may share memory.
-   * build gasneti_nodemap[]
-   *  call gasneti_nodemapParse() after constructing it.
-   */
-  gasneti_nodemap = gasneti_malloc(gasneti_nodes * sizeof(uint32_t));
-  gasneti_assert(gasneti_nodemap);
-  /* PMI uses int, gni and gasnet use uint32_t */
-  gasneti_assert(sizeof(int32_t) == sizeof(gasnett_atomic_t));
-  gasneti_assert(sizeof(int) == sizeof(uint32_t));
-  ret = PMI_Get_numpes_on_smp(&localranks);
-  gasneti_assert(ret == PMI_SUCCESS);
-  gasneti_assert(localranks <= gasneti_nodes);
-  /* OK to use the base of gasneti_nodemap as a temp because it isn't filled in */
-  ret = PMI_Get_pes_on_smp((int *) gasneti_nodemap, localranks);
-  gasneti_assert(ret == PMI_SUCCESS);
-  /* find minimum rank on local supernode */
-  minlocalrank = gasneti_nodes;  /* one larger than largest possible */
-  for (i = 0; i < localranks; i += 1) {
-    if (gasneti_nodemap[i] < minlocalrank) minlocalrank = gasneti_nodemap[i];
+  /* Retreive the nidlist to construct the gasneti_nodemap[]  */
+  { int *nidlist;
+    ret = PMI_Get_nidlist_ptr((void**) &nidlist);
+    gasneti_assert(ret == PMI_SUCCESS);
+    gasneti_nodemapInit(NULL, nidlist, sizeof(int), sizeof(int));
   }
-  gasnetc_bootstrapExchange(&minlocalrank, sizeof(uint32_t), gasneti_nodemap);
-  for (i = 0; i < gasneti_nodes; i += 1) {
-    /* gasneti_assert(gasneti_nodemap[i] >= 0);  type is unsigned, so this is moot */
-    gasneti_assert(gasneti_nodemap[i] < gasneti_nodes);
-  }
-
-  gasneti_nodemapParse();
 
   #if GASNET_PSHM
     /* If your conduit will support PSHM, you should initialize it here.
-     * The 1st argument is normally "&gasnetc_bootstrapExchange" (described below).
+     * The 1st argument is normally "&gasnetc_bootstrapSNodeBroadcast" or equivalent
      * The 2nd argument is the amount of shared memory space needed for any
      * conduit-specific uses.  The return value is a pointer to the space
      * requested by the 2nd argument.
      */
-    gasnetc_exitcodes = gasneti_pshm_init(&gasnetc_bootstrapExchange,
+    gasnetc_exitcodes = gasneti_pshm_init(&gasneti_bootstrapSNodeBroadcast_pmi,
                                           gasneti_nodemap_local_count * sizeof(gasnetc_exitcode_t));
     gasnetc_exitcodes[gasneti_nodemap_local_rank].present = 0;
   #endif
@@ -644,7 +645,7 @@ static int gasnetc_init(int *argc, char ***argv) {
       /* localSegmentLimit provides a conduit-specific limit on the max segment size.
        * can use (uintptr_t)-1 as unlimited.
        */
-      gasneti_segmentInit( max_pin, &gasnetc_bootstrapExchange);
+      gasneti_segmentInit( max_pin, &gasnetc_bootstrapExchange_gni);
 
 
     }
@@ -687,7 +688,10 @@ extern int gasnet_init(int *argc, char ***argv) {
   /* after this, ams should work, but the segments aren't registered yet */
   int retval = gasnetc_init(argc, argv);
   if (retval != GASNET_OK) GASNETI_RETURN(retval);
+  #if 0
+  /* Already done in gasnetc_init() to allow tracing of init steps */
   gasneti_trace_init(argc, argv);
+  #endif
   return GASNET_OK;
 }
 /* ------------------------------------------------------------------------------------ */
@@ -743,6 +747,7 @@ extern int gasnetc_attach(gasnet_handlerentry_t *table, int numentries,
                           uintptr_t segsize, uintptr_t minheapoffset) {
   GASNETC_DIDX_POST(GASNETC_DEFAULT_DOMAIN);
   void *segbase = NULL;
+  int numreg = 0;
   
   GASNETI_TRACE_PRINTF(C,("gasnetc_attach(table (%i entries), segsize=%lu, minheapoffset=%lu)",
                           numentries, (unsigned long)segsize, (unsigned long)minheapoffset));
@@ -776,10 +781,9 @@ extern int gasnetc_attach(gasnet_handlerentry_t *table, int numentries,
   { /*  core API handlers */
     gasnet_handlerentry_t *ctable = (gasnet_handlerentry_t *)gasnetc_get_handlertable();
     int len = 0;
-    int numreg = 0;
     gasneti_assert(ctable);
     while (ctable[len].fnptr) len++; /* calc len */
-    if (gasnetc_reghandlers(ctable, len, 1, 63, 0, &numreg) != GASNET_OK)
+    if (gasnetc_reghandlers(ctable, len, 1, GASNETE_HANDLER_BASE-1, 0, &numreg) != GASNET_OK)
       GASNETI_RETURN_ERRR(RESOURCE,"Error registering core API handlers");
     gasneti_assert(numreg == len);
   }
@@ -787,14 +791,27 @@ extern int gasnetc_attach(gasnet_handlerentry_t *table, int numentries,
   { /*  extended API handlers */
     gasnet_handlerentry_t *etable = (gasnet_handlerentry_t *)gasnete_get_handlertable();
     int len = 0;
-    int numreg = 0;
     gasneti_assert(etable);
     while (etable[len].fnptr) len++; /* calc len */
-    if (gasnetc_reghandlers(etable, len, 64, 127, 0, &numreg) != GASNET_OK)
+    if (gasnetc_reghandlers(etable, len, GASNETE_HANDLER_BASE, 127, 0, &numreg) != GASNET_OK)
       GASNETI_RETURN_ERRR(RESOURCE,"Error registering extended API handlers");
     gasneti_assert(numreg == len);
   }
 
+#if GASNETC_GNI_FIREHOSE
+  { /* firehose handlers */
+    gasnet_handlerentry_t *ftable = (gasnet_handlerentry_t *)firehose_get_handlertable();
+    int len = 0;
+    int base = GASNETE_HANDLER_BASE + numreg;   /* start right after etable */
+    gasneti_assert(ftable);
+    while (ftable[len].fnptr) len++; /* calc len */
+    gasneti_assert(base + len <= 128);  /* enough space remaining after etable? */
+    if (gasnetc_reghandlers(ftable, len, base, 127, 1, &numreg) != GASNET_OK)
+      GASNETI_RETURN_ERRR(RESOURCE, "Error registering firehose handlers");
+    gasneti_assert(numreg == len);
+  }
+#endif
+
   if (table) { /*  client handlers */
     int numreg1 = 0;
     int numreg2 = 0;
@@ -837,7 +854,7 @@ extern int gasnetc_attach(gasnet_handlerentry_t *table, int numentries,
   #if GASNET_SEGMENT_FAST || GASNET_SEGMENT_LARGE
     if (segsize == 0) segbase = NULL; /* no segment */
     else {
-      gasneti_segmentAttach(segsize, minheapoffset, gasneti_seginfo, &gasnetc_bootstrapExchange);
+      gasneti_segmentAttach(segsize, minheapoffset, gasneti_seginfo, &gasnetc_bootstrapExchange_gni);
       segbase = gasneti_seginfo[gasneti_mynode].addr;
       segsize = gasneti_seginfo[gasneti_mynode].size;
       gasnetc_assert_aligned(segbase, GASNET_PAGESIZE);
@@ -869,10 +886,39 @@ extern int gasnetc_attach(gasnet_handlerentry_t *table, int numentries,
   /* (LCS) This was done by segmentAttach above
    */
 
+  /* ------------------------------------------------------------------------------------ */
+  /* Initialize firehose */
+#if GASNETC_GNI_FIREHOSE
+  gasnetc_use_firehose = gasneti_getenv_yesno_withdefault("GASNET_USE_FIREHOSE", 1);
+  if (gasnetc_use_firehose && gasneti_nodes > 1) {
+    /* Configure firehose with capacity to pin the out-of-segment memory twice over */
+    uintptr_t pm_limit = gasneti_getPhysMemSz(1) *
+                         gasneti_getenv_dbl_withdefault("GASNET_PHYSMEM_PINNABLE_RATIO",
+                                                        GASNETC_DEFAULT_PHYSMEM_PINNABLE_RATIO);
+    uintptr_t firehose_mem = 2 * (pm_limit - segsize);
+    size_t firehose_reg = INT_MAX; /* Will be reduced by firehose_init */
+    size_t max_pinsize = 0x200000; /* 2M default - env var can override */
+    int flags = FIREHOSE_INIT_FLAG_LOCAL_ONLY;
+
+    /* TODO: add in the prepinned buffers (IMM and AM) ? */
+    firehose_init(firehose_mem, firehose_reg, max_pinsize,
+                  NULL, 0, flags, &gasnetc_firehose_info);
+    gasneti_assert(gasnetc_firehose_info.max_LocalPinSize >= GASNETC_MAX_LONG + FH_BUCKET_SIZE);
+
+    /* Determine alignment (and max size) for fh requests - a power-of-two <= max_region/2 */
+    gasnetc_fh_align = GASNET_PAGESIZE;
+    while ((gasnetc_fh_align * 2) <= (gasnetc_firehose_info.max_LocalPinSize / 2)) {
+      gasnetc_fh_align *= 2;
+    }
+    gasnetc_fh_align_mask = gasnetc_fh_align - 1;
+    gasnetc_did_firehose_init = 1;
+  }
+#endif
+
   /* ------------------------------------------------------------------------------------ */
   /*  primary attach complete */
   gasneti_attach_done = 1;
-  gasnetc_bootstrapBarrier();
+  gasnetc_bootstrapBarrier_gni();
 
   GASNETI_TRACE_PRINTF(C,("gasnetc_attach(): primary attach complete"));
 
@@ -891,13 +937,18 @@ extern int gasnetc_attach(gasnet_handlerentry_t *table, int numentries,
   gasneti_nodemapFini();
 
   /* ensure extended API is initialized across nodes */
-  gasnetc_bootstrapBarrier();
+  gasnetc_bootstrapBarrier_gni();
   gasnetc_sys_coll_fini();
 
   GASNETI_TRACE_PRINTF(C,("gasnetc_attach: done\n"));
   return GASNET_OK;
 }
 /* ------------------------------------------------------------------------------------ */
+static int gasnetc_exit_in_signal = 0;  /* to avoid certain things in signal context */
+extern void gasnetc_fatalsignal_callback(int sig) {
+  gasnetc_exit_in_signal = 1;
+}
+
 static int gasnetc_remoteShutdown = 0;
 
 #if HAVE_ON_EXIT
@@ -1028,6 +1079,15 @@ extern void gasnetc_exit(int exitcode) {
   }
   alarm(0);
 
+#if GASNETC_GNI_FIREHOSE
+  if (gasnetc_did_firehose_init && !gasnetc_exit_in_signal) {
+    /* Note we skip firehose_fini() on exit via a signal */
+    alarm(10);
+    firehose_fini();
+    alarm(0);
+  }
+#endif
+
   gasneti_flush_streams();
   gasneti_trace_finish();
   gasneti_sched_yield();
@@ -1112,8 +1172,8 @@ extern int gasnetc_AMPoll(void)
   ================================
 */
 
-GASNETI_INLINE(gasnetc_general_am_send) 
-int gasnetc_general_am_send(gasnetc_post_descriptor_t *gpd)
+GASNETI_INLINE(gasnetc_general_am_send_common)
+int gasnetc_general_am_send_common(gasnetc_post_descriptor_t *gpd)
 {
   int retval;
   gasneti_suspend_spinpollers();
@@ -1122,6 +1182,16 @@ int gasnetc_general_am_send(gasnetc_post_descriptor_t *gpd)
   return(retval);
 }
 
+#define gasnetc_general_am_send_request gasnetc_general_am_send_common
+
+GASNETI_INLINE(gasnetc_general_am_send_reply)
+int gasnetc_general_am_send_reply(gasnetc_post_descriptor_t *gpd, gasnet_token_t t)
+{
+  return ((gasnetc_token_t *)t)->deferred_reply
+             ? GASNET_OK
+             : gasnetc_general_am_send_common(gpd);
+}
+
 /*------------------- local delivery cases (non-PSHM) ------------------ */
 #if !GASNET_PSHM
 
@@ -1132,7 +1202,7 @@ int gasnetc_local_short_common(int is_req, gasnet_handler_t handler,
   int i;
   
   const gasneti_handler_fn_t handler_fn = gasnetc_handler[handler];
-  gasnetc_token_t the_token = { gasneti_mynode, is_req, 0 };
+  gasnetc_token_t the_token = { gasneti_mynode, is_req, 0, NULL };
   gasnet_token_t token = (gasnet_token_t)&the_token; /* RUN macros need an lvalue */
   gasnet_handlerarg_t args[gasnet_AMMaxArgs()];
   
@@ -1151,7 +1221,7 @@ int gasnetc_local_medium_common(int is_req, gasnet_handler_t handler,
   int i;
   
   const gasneti_handler_fn_t handler_fn = gasnetc_handler[handler];
-  gasnetc_token_t the_token = { gasneti_mynode, is_req, 0 };
+  gasnetc_token_t the_token = { gasneti_mynode, is_req, 0, NULL };
   gasnet_token_t token = (gasnet_token_t)&the_token; /* RUN macros need an lvalue */
   gasnet_handlerarg_t args[gasnet_AMMaxArgs()];
   void *payload = alloca(nbytes);
@@ -1172,7 +1242,7 @@ int gasnetc_local_long_common(int is_req, gasnet_handler_t handler,
                                int numargs, va_list argptr)
 {
   const gasneti_handler_fn_t handler_fn = gasnetc_handler[handler];
-  gasnetc_token_t the_token = { gasneti_mynode, is_req, 0 };
+  gasnetc_token_t the_token = { gasneti_mynode, is_req, 0, NULL };
   gasnet_token_t token = (gasnet_token_t)&the_token; /* RUN macros need an lvalue */
   gasnet_handlerarg_t args[gasnet_AMMaxArgs()];
   int i;
@@ -1275,6 +1345,11 @@ int gasnetc_put_long_payload( gasnet_node_t dest,
     gasnetc_post_descriptor_t *gpd = gasnetc_alloc_post_descriptor(GASNETC_DIDX_PASS_ALONE);
     gpd->gpd_completion = (uintptr_t) completed_p;
     gpd->flags = GC_POST_COMPLETION_CNTR;
+#if GASNETC_GNI_FIREHOSE
+    if (gasnetc_use_firehose && !gasneti_in_segment(gasneti_mynode, src_addr, chunk)) {
+      chunk= gasnetc_rdma_put_fh(dest, dst_addr, src_addr, chunk, gpd);
+    } else
+#endif
     chunk = gasnetc_rdma_put_bulk(dest, dst_addr, src_addr, chunk, gpd);
     if_pt (0 == (nbytes -= chunk)) break; /* expect to finish in one pass */
 
@@ -1305,6 +1380,11 @@ int gasnetc_put_longasync_payload( gasnet_node_t dest,
     gasnetc_post_descriptor_t *gpd = gasnetc_alloc_post_descriptor(GASNETC_DIDX_PASS_ALONE);
     gpd->gpd_completion = (uintptr_t) header_gpd;
     gpd->flags = GC_POST_COMPLETION_SEND;
+#if GASNETC_GNI_FIREHOSE
+    if (gasnetc_use_firehose && !gasneti_in_segment(gasneti_mynode, src_addr, chunk)) {
+      chunk = gasnetc_rdma_put_fh(dest, dst_addr, src_addr, chunk, gpd);
+    } else
+#endif
     chunk = gasnetc_rdma_put_bulk(dest, dst_addr, src_addr, chunk, gpd);
     if_pt (0 == (nbytes -= chunk)) break; /* expect to finish in one pass */
 
@@ -1346,7 +1426,7 @@ extern int gasnetc_AMRequestShortM(
     const size_t total_len = GASNETC_HEADLEN(short, numargs);
     gasnetc_post_descriptor_t *gpd = gasnetc_alloc_request_post_descriptor(dest, total_len);
     gasnetc_format_short(gpd, handler, numargs, argptr);
-    retval = gasnetc_general_am_send(gpd);
+    retval = gasnetc_general_am_send_request(gpd);
   }
   va_end(argptr);
   GASNETI_RETURN(retval);
@@ -1378,7 +1458,7 @@ extern int gasnetc_AMRequestMediumM(
     const size_t total_len = GASNETC_HEADLEN(medium, numargs) + nbytes;
     gasnetc_post_descriptor_t *gpd = gasnetc_alloc_request_post_descriptor(dest, total_len);
     gasnetc_format_medium(gpd, handler,source_addr,nbytes,numargs,argptr);
-    retval = gasnetc_general_am_send(gpd);
+    retval = gasnetc_general_am_send_request(gpd);
   }
   va_end(argptr);
   GASNETI_RETURN(retval);
@@ -1431,7 +1511,7 @@ extern int gasnetc_AMRequestLongM( gasnet_node_t dest,        /* destination nod
       /* Poll for the RDMA completion */
       gasnetc_wait_long_payload(initiated, &completed GASNETC_DIDX_PASS);
     }
-    retval = gasnetc_general_am_send(gpd);
+    retval = gasnetc_general_am_send_request(gpd);
   }
   va_end(argptr);
   GASNETI_RETURN(retval);
@@ -1468,7 +1548,7 @@ extern int gasnetc_AMRequestLongAsyncM( gasnet_node_t dest,        /* destinatio
     gasnetc_format_long(gpd, is_packed, handler, nbytes, dest_addr, numargs, argptr);
     if (is_packed) {
       memcpy((void*)(gpd->gpd_am_packet + head_len), source_addr, nbytes);
-      retval = gasnetc_general_am_send(gpd);
+      retval = gasnetc_general_am_send_request(gpd);
     } else {
       /* Rdma data, then send header as part of completion*/
       GASNETC_DIDX_POST((gasnete_mythread())->domain_idx);
@@ -1513,7 +1593,7 @@ extern int gasnetc_AMReplyShortM(
     const size_t total_len = GASNETC_HEADLEN(short, numargs);
     gasnetc_post_descriptor_t *gpd = gasnetc_alloc_reply_post_descriptor(token, total_len);
     gasnetc_format_short(gpd, handler,numargs,argptr);
-    retval = gasnetc_general_am_send(gpd);
+    retval = gasnetc_general_am_send_reply(gpd, token);
   }
   va_end(argptr);
   GASNETI_RETURN(retval);
@@ -1545,7 +1625,7 @@ extern int gasnetc_AMReplyMediumM(
     const size_t total_len = GASNETC_HEADLEN(medium, numargs) + nbytes;
     gasnetc_post_descriptor_t *gpd = gasnetc_alloc_reply_post_descriptor(token, total_len);
     gasnetc_format_medium(gpd, handler,source_addr,nbytes,numargs,argptr);
-    retval = gasnetc_general_am_send(gpd);
+    retval = gasnetc_general_am_send_reply(gpd, token);
   }
   va_end(argptr);
   GASNETI_RETURN(retval);
@@ -1597,7 +1677,7 @@ extern int gasnetc_AMReplyLongM(
       /* Poll for the RDMA completion */
       gasnetc_wait_long_payload(initiated, &completed GASNETC_DIDX_PASS);
     }
-    retval = gasnetc_general_am_send(gpd);
+    retval = gasnetc_general_am_send_reply(gpd, token);
   }
   va_end(argptr);
   GASNETI_RETURN(retval);
