@@ -676,9 +676,10 @@ static void gasnete_ibdbarrier_init(gasnete_coll_team_t team);
     gasnete_coll_default_barrier_type = GASNETE_COLL_BARRIER_IBDISSEM;  \
 } while (0)
 
-#define GASNETE_BARRIER_INIT(TEAM, BARRIER_TYPE) do {                                  \
-    if ((BARRIER_TYPE) == GASNETE_COLL_BARRIER_IBDISSEM && (TEAM)==GASNET_TEAM_ALL) { \
-      gasnete_ibdbarrier_init(TEAM);                              \
+#define GASNETE_BARRIER_INIT(TEAM, TYPE, NODES, SUPERNODES) do { \
+    if ((TYPE) == GASNETE_COLL_BARRIER_IBDISSEM &&               \
+        (TEAM) == GASNET_TEAM_ALL) {                             \
+      gasnete_ibdbarrier_init(TEAM);                             \
     }                                                            \
   } while (0)
 
@@ -696,7 +697,8 @@ static int gasnete_conduit_rdmabarrier(const char *barrier, gasneti_auxseg_reque
     /* TODO: could keep the full space and allocate some to additional teams */
     size_t request;
 #if GASNETI_PSHM_BARRIER_HIER
-    const int size = gasneti_nodemap_global_count;
+    const int is_hier = gasneti_getenv_yesno_withdefault("GASNET_PSHM_BARRIER_HIER", 1);
+    const int size = is_hier ? gasneti_nodemap_global_count : gasneti_nodes;
 #else
     const int size = gasneti_nodes;
 #endif
@@ -751,9 +753,8 @@ typedef struct {
   gasnete_pshmbarrier_data_t *barrier_pshm; /* non-NULL if using hierarchical code */
   int barrier_passive;        /*  2 if some other node makes progress for me, 0 otherwise */
 #endif
-  int barrier_size;           /*  ceil(lg(nodes)) */
-  int barrier_goal;           /*  (ceil(lg(nodes)) << 1) */
-  int volatile barrier_slot;  /*  (step << 1) | phase */
+  int barrier_goal;           /*  (1+ceil(lg(nodes)) << 1) == final barrier_state for phase=0 */
+  int volatile barrier_state; /*  (step << 1) | phase, where step is 1-based (0 is pshm notify) */
   int volatile barrier_value; /*  barrier value (evolves from local value) */
   int volatile barrier_flags; /*  barrier flags (evolves from local value) */
   volatile uint64_t *barrier_inbox; /*  in-segment memory to recv notifications */
@@ -762,7 +763,7 @@ typedef struct {
 /* Unlike the extended-ref version we CAN assume that a 64-bit write
  * will be atomic, and so can use a "present" bit in flags.
  */
-#define GASNETE_IBDBARRIER_PRESENT 0x80000000
+#define GASNETE_IBDBARRIER_PRESENT GASNETE_BARRIERFLAG_CONDUIT0
 #define GASNETE_IBDBARRIER_VALUE(_u64) GASNETI_HIWORD(_u64)
 #define GASNETE_IBDBARRIER_FLAGS(_u64) GASNETI_LOWORD(_u64)
 #define GASNETE_IBDBARRIER_BUILD(_value,_flags) \
@@ -770,67 +771,125 @@ typedef struct {
   
 /* Pad struct to a specfic size and interleave */
 #define GASNETE_IBDBARRIER_INBOX_WORDS (GASNETE_RDMABARRIER_INBOX_SZ/sizeof(uint64_t))
-#define GASNETE_IBDBARRIER_INBOX(_bd,_slot)     \
+#define GASNETE_IBDBARRIER_INBOX(_bd,_state)     \
             (((_bd)->barrier_inbox) \
-                       + (unsigned)(_slot) * GASNETE_IBDBARRIER_INBOX_WORDS)
-#define GASNETE_IBDBARRIER_INBOX_REMOTE(_bd,_step,_slot)  \
+                       + (unsigned)((_state)-2) * GASNETE_IBDBARRIER_INBOX_WORDS)
+#define GASNETE_IBDBARRIER_INBOX_REMOTE(_bd,_step,_state)  \
             (((_bd)->barrier_peers[(unsigned)(_step)].addr \
-                       + (unsigned)(_slot) * GASNETE_IBDBARRIER_INBOX_WORDS))
+                       + (unsigned)((_state)-2) * GASNETE_IBDBARRIER_INBOX_WORDS))
 #define GASNETE_IBDBARRIER_INBOX_NEXT(_addr)    \
             ((_addr) + 2U * GASNETE_IBDBARRIER_INBOX_WORDS)
 
 GASNETI_INLINE(gasnete_ibdbarrier_send)
 void gasnete_ibdbarrier_send(gasnete_coll_ibdbarrier_t *barrier_data,
-                             int numsteps, unsigned int slot,
+                             int numsteps, unsigned int state,
                              gasnet_handlerarg_t value, gasnet_handlerarg_t flags) {
   GASNETE_THREAD_LOOKUP
-  unsigned int step = slot >> 1;
+  unsigned int step = state >> 1;
   int i;
 
   /* Use the upper half (padding) an "other phase" inbox as an in-segment temporary.
    * This has sufficient lifetime for bulk and sufficient alignment for non-bulk.
    * Use of opposite phase prevents cacheline contention with arrivals.
    */
-  volatile uint64_t *payload = GASNETE_IBDBARRIER_INBOX(barrier_data, (slot^1))
+  const uint64_t msg = GASNETE_IBDBARRIER_BUILD(value, flags);
+  volatile uint64_t *payload = GASNETE_IBDBARRIER_INBOX(barrier_data, (state^1))
                                     + (GASNETE_IBDBARRIER_INBOX_WORDS/2);
-  *payload = GASNETE_IBDBARRIER_BUILD(value, flags);
+  *payload = msg;
 
   /* TODO:
    *   Reduce latency by pre-computing sr_desc and rkey_index at init time
    *   AND/OR providing a specialized alternative to gasnetc_rdma_put()
    */
+  gasneti_assert(state < barrier_data->barrier_goal);
 
-  for (i = 0; i < numsteps; ++i, slot += 2, step += 1) {
+  for (i = 0; i < numsteps; ++i, state += 2, step += 1) {
     const gasnet_node_t node = barrier_data->barrier_peers[step].node;
-    uint64_t * const dst = GASNETE_IBDBARRIER_INBOX_REMOTE(barrier_data, step, slot);
+    uint64_t * const dst = GASNETE_IBDBARRIER_INBOX_REMOTE(barrier_data, step, state);
+#if GASNET_PSHM
+    if (gasneti_pshm_in_supernode(node)) {
+      *(uint64_t*)gasneti_pshm_addr2local(node, dst) = msg;
+    } else
+#endif
     (void) gasnetc_rdma_put(node, (void*)payload, dst, sizeof(*payload), NULL, NULL, NULL GASNETE_THREAD_PASS);
   }
 }
 
+#if GASNETI_PSHM_BARRIER_HIER
+static int gasnete_ibdbarrier_kick_pshm(gasnete_coll_team_t team) {
+  gasnete_coll_ibdbarrier_t *barrier_data = team->barrier_data;
+  int done = (barrier_data->barrier_state > 1);
+
+  if (!done && !gasnete_ibdbarrier_trylock(&barrier_data->barrier_lock)) {
+    const int state = barrier_data->barrier_state;
+    done = (state > 1);
+    if (!done) {
+      PSHM_BDATA_DECL(pshm_bdata, barrier_data->barrier_pshm);
+      if (gasnete_pshmbarrier_kick(pshm_bdata)) {
+        const int value = pshm_bdata->shared->value;
+        const int flags = pshm_bdata->shared->flags;
+        barrier_data->barrier_value = value;
+        barrier_data->barrier_flags = flags;
+        gasneti_sync_writes();
+        barrier_data->barrier_state = state + 2;
+        gasnete_ibdbarrier_unlock(&barrier_data->barrier_lock); /* Cannot send while holding HSL */
+        if ((barrier_data->barrier_goal > 2) && !barrier_data->barrier_passive) {
+          gasnete_ibdbarrier_send(barrier_data, 1, state+2, value, flags);
+        } else {
+          gasnete_barrier_pf_disable(team);
+        }
+        return 1;
+      }
+    }
+    gasnete_ibdbarrier_unlock(&barrier_data->barrier_lock);
+  }
+
+  return done;
+}
+#endif
+
 void gasnete_ibdbarrier_kick(gasnete_coll_team_t team) {
   gasnete_coll_ibdbarrier_t *barrier_data = team->barrier_data;
   volatile uint64_t *inbox;
   uint64_t result;
   int numsteps = 0;
-  int slot, cursor;
+  int state, new_state;
   int flags, value;
 
   /* early unlocked read: */
-  slot = barrier_data->barrier_slot;
-
-  if (slot >= barrier_data->barrier_goal)
+  state = barrier_data->barrier_state;
+  if (state >= barrier_data->barrier_goal)
     return; /* nothing to do */
 
-  gasneti_assert(team->total_ranks > 1); /* singleton should have matched (slot >= goal), above */
+  gasneti_assert(team->total_ranks > 1); /* singleton should have matched (state >= goal), above */
+
+#if GASNETI_PSHM_BARRIER_HIER
+  if (barrier_data->barrier_pshm) {
+    /* Cannot begin to probe until local notify is complete */
+    if (!gasnete_ibdbarrier_kick_pshm(team)) return;
+  }
+#endif
 
-#if GASNETI_THREADS
   if (gasnete_ibdbarrier_trylock(&barrier_data->barrier_lock))
     return; /* another thread is currently in kick */
 
-  /* reread w/ lock held: */
-  slot = barrier_data->barrier_slot;
+  /* reread w/ lock held and/or because kick_pshm may have advanced it */
+  state = barrier_data->barrier_state;
 
-  if_pf (slot < 2) {/* need to pick up value/flags from notify */
+#if GASNETI_PSHM_BARRIER_HIER
+  if_pf (state < 2) { /* local notify has not completed */
+    gasnete_ibdbarrier_unlock(&barrier_data->barrier_lock);
+    return;
+  } else if (barrier_data->barrier_passive) {
+    gasnete_barrier_pf_disable(team);
+    gasnete_ibdbarrier_unlock(&barrier_data->barrier_lock);
+    return;
+  }
+  gasneti_assert(!barrier_data->barrier_passive);
+#endif
+
+#if GASNETI_THREADS
+  if_pf (state < 4) {/* need to pick up value/flags from notify */
     gasneti_sync_reads(); /* value/flags were written by the non-locked notify */
   }
 #endif
@@ -839,8 +898,8 @@ void gasnete_ibdbarrier_kick(gasnete_coll_team_t team) {
   flags = barrier_data->barrier_flags;
 
   /* process all consecutive steps which have arrived since we last ran */
-  inbox = GASNETE_IBDBARRIER_INBOX(barrier_data, slot);
-  for (cursor = slot; cursor < barrier_data->barrier_goal && (0 != (result = *inbox)); cursor+=2) {
+  inbox = GASNETE_IBDBARRIER_INBOX(barrier_data, state);
+  for (new_state = state; new_state < barrier_data->barrier_goal && (0 != (result = *inbox)); new_state+=2) {
     const int step_value = GASNETE_IBDBARRIER_VALUE(result);
     const int step_flags = GASNETE_IBDBARRIER_FLAGS(result);
     *inbox = 0;
@@ -862,53 +921,55 @@ void gasnete_ibdbarrier_kick(gasnete_coll_team_t team) {
     barrier_data->barrier_flags = flags; 
     barrier_data->barrier_value = value; 
 
-    if (cursor >= barrier_data->barrier_goal) { /* We got the last recv - barrier locally complete */
+    if (new_state >= barrier_data->barrier_goal) { /* We got the last recv - barrier locally complete */
       gasnete_barrier_pf_disable(team);
-      gasneti_sync_writes(); /* flush state before the write to barrier_slot below */
+      gasneti_sync_writes(); /* flush state before the write to barrier_state below */
       numsteps -= 1; /* no send at last step */
     } 
     /* notify all threads of the step increase - 
        this may allow other local threads to proceed on the barrier and even indicate
        barrier completion while we overlap outgoing notifications to other nodes
     */
-    barrier_data->barrier_slot = cursor;
+    barrier_data->barrier_state = new_state;
   } 
 
   gasnete_ibdbarrier_unlock(&barrier_data->barrier_lock);
 
   if (numsteps) { /* need to issue one or more Puts */
-    gasnete_ibdbarrier_send(barrier_data, numsteps, slot+2, value, flags);
+    gasnete_ibdbarrier_send(barrier_data, numsteps, state+2, value, flags);
   }
 }
 
 static void gasnete_ibdbarrier_notify(gasnete_coll_team_t team, int id, int flags) {
   gasnete_coll_ibdbarrier_t *barrier_data = team->barrier_data;
+  int state = 2 + ((barrier_data->barrier_state & 1) ^ 1); /* enter new phase */
   int do_send = 1;
-  int slot;
+  int want_pf = 1;
 
   GASNETE_SPLITSTATE_NOTIFY_ENTER(team);
 
 #if GASNETI_PSHM_BARRIER_HIER
   if (barrier_data->barrier_pshm) {
     PSHM_BDATA_DECL(pshm_bdata, barrier_data->barrier_pshm);
-    (void)gasnete_pshmbarrier_notify_inner(pshm_bdata, id, flags);
-    do_send = !barrier_data->barrier_passive;
-    id = pshm_bdata->shared->value;
-    flags = pshm_bdata->shared->flags;
+    if (gasnete_pshmbarrier_notify_inner(pshm_bdata, id, flags)) {
+      id = pshm_bdata->shared->value;
+      flags = pshm_bdata->shared->flags;
+      want_pf = do_send = !barrier_data->barrier_passive;
+    } else {
+      do_send = 0;
+      state -= 2;
+    }
   }
 #endif
 
   barrier_data->barrier_value = id;
   barrier_data->barrier_flags = flags;
 
-  slot = ((barrier_data->barrier_slot & 1) ^ 1); /* enter new phase */
   gasneti_sync_writes();
-  barrier_data->barrier_slot = slot;
+  barrier_data->barrier_state = state;
 
-  if (do_send) {
-    gasnete_ibdbarrier_send(barrier_data, 1, slot, id, flags);
-    gasnete_barrier_pf_enable(team);
-  }
+  if (do_send) gasnete_ibdbarrier_send(barrier_data, 1, state, id, flags);
+  if (want_pf) gasnete_barrier_pf_enable(team);
 
   /*  update state */
   gasneti_sync_writes(); /* ensure all state changes committed before return */
@@ -917,21 +978,33 @@ static void gasnete_ibdbarrier_notify(gasnete_coll_team_t team, int id, int flag
 /* Notify specialized to one (super)node case (reduced branches in BOTH variants) */
 static void gasnete_ibdbarrier_notify_singleton(gasnete_coll_team_t team, int id, int flags) {
   gasnete_coll_ibdbarrier_t *barrier_data = team->barrier_data;
+#if GASNETI_PSHM_BARRIER_HIER
+  int state = 2;
+#endif
 
   GASNETE_SPLITSTATE_NOTIFY_ENTER(team);
 
 #if GASNETI_PSHM_BARRIER_HIER
   if (barrier_data->barrier_pshm) {
     PSHM_BDATA_DECL(pshm_bdata, barrier_data->barrier_pshm);
-    (void)gasnete_pshmbarrier_notify_inner(pshm_bdata, id, flags);
-    id = pshm_bdata->shared->value;
-    flags = pshm_bdata->shared->flags;
+    if (gasnete_pshmbarrier_notify_inner(pshm_bdata, id, flags)) {
+      id = pshm_bdata->shared->value;
+      flags = pshm_bdata->shared->flags;
+    } else {
+      state = 0;
+    }
   }
 #endif
 
   barrier_data->barrier_value = id;
   barrier_data->barrier_flags = flags;
 
+#if GASNETI_PSHM_BARRIER_HIER
+  gasneti_sync_writes();
+  barrier_data->barrier_state = state;
+  if (!state) gasnete_barrier_pf_enable(team);
+#endif
+
   /*  update state */
   gasneti_sync_writes(); /* ensure all state changes committed before return */
 }
@@ -939,7 +1012,7 @@ static void gasnete_ibdbarrier_notify_singleton(gasnete_coll_team_t team, int id
 static int gasnete_ibdbarrier_wait(gasnete_coll_team_t team, int id, int flags) {
   gasnete_coll_ibdbarrier_t *barrier_data = team->barrier_data;
 #if GASNETI_PSHM_BARRIER_HIER
-  const PSHM_BDATA_DECL(pshm_bdata, barrier_data->barrier_pshm);
+  PSHM_BDATA_DECL(pshm_bdata, barrier_data->barrier_pshm);
 #endif
   int retval = GASNET_OK;
 
@@ -949,6 +1022,7 @@ static int gasnete_ibdbarrier_wait(gasnete_coll_team_t team, int id, int flags)
 #if GASNETI_PSHM_BARRIER_HIER
   if (pshm_bdata) {
     const int passive_shift = barrier_data->barrier_passive;
+    gasneti_polluntil(gasnete_ibdbarrier_kick_pshm(team));
     retval = gasnete_pshmbarrier_wait_inner(pshm_bdata, id, flags, passive_shift);
     if (passive_shift) {
       /* Once the active peer signals done, we can return */
@@ -960,14 +1034,14 @@ static int gasnete_ibdbarrier_wait(gasnete_coll_team_t team, int id, int flags)
   }
 #endif
 
-  if (barrier_data->barrier_slot >= barrier_data->barrier_goal) {
+  if (barrier_data->barrier_state >= barrier_data->barrier_goal) {
     /* completed asynchronously before wait (via progressfns or try) */
     GASNETI_TRACE_EVENT_TIME(B,BARRIER_ASYNC_COMPLETION,GASNETI_TICKS_NOW_IFENABLED(B)-gasnete_barrier_notifytime);
   } else {
     /* kick once, and if still necessary, wait for a response */
     gasnete_ibdbarrier_kick(team);
     /* cannot BLOCKUNTIL since progess may occur on non-AM events */
-    while (barrier_data->barrier_slot < barrier_data->barrier_goal) {
+    while (barrier_data->barrier_state < barrier_data->barrier_goal) {
       GASNETI_WAITHOOK();
       GASNETI_SAFE(gasneti_AMPoll());
       gasnete_ibdbarrier_kick(team);
@@ -1011,7 +1085,8 @@ static int gasnete_ibdbarrier_try(gasnete_coll_team_t team, int id, int flags) {
 #if GASNETI_PSHM_BARRIER_HIER
   if (barrier_data->barrier_pshm) {
     const int passive_shift = barrier_data->barrier_passive;
-    if (!gasnete_pshmbarrier_try_inner(barrier_data->barrier_pshm, passive_shift))
+    if (!gasnete_ibdbarrier_kick_pshm(team) ||
+        !gasnete_pshmbarrier_try_inner(barrier_data->barrier_pshm, passive_shift))
       return GASNET_ERR_NOT_READY;
     if (passive_shift)
       return gasnete_ibdbarrier_wait(team, id, flags);
@@ -1020,7 +1095,8 @@ static int gasnete_ibdbarrier_try(gasnete_coll_team_t team, int id, int flags) {
 #endif
     gasnete_ibdbarrier_kick(team);
 
-  if (barrier_data->barrier_slot >= barrier_data->barrier_goal) return gasnete_ibdbarrier_wait(team, id, flags);
+  if (barrier_data->barrier_state >= barrier_data->barrier_goal)
+    return gasnete_ibdbarrier_wait(team, id, flags);
   else return GASNET_ERR_NOT_READY;
 }
 
@@ -1063,11 +1139,11 @@ static void gasnete_ibdbarrier_init(gasnete_coll_team_t team) {
   int steps;
   int total_ranks = team->total_ranks;
   int myrank = team->myrank;
+  gasnete_coll_peer_list_t *peers = &team->peers;
+
 #if GASNETI_PSHM_BARRIER_HIER
-  gasnet_node_t *supernode_reps = NULL;
-  PSHM_BDATA_DECL(pshm_bdata, gasnete_pshmbarrier_init_hier(team, &total_ranks, &myrank, &supernode_reps));
+  PSHM_BDATA_DECL(pshm_bdata, gasnete_pshmbarrier_init_hier(team, &total_ranks, &myrank, &peers));
 #endif
-  int64_t j;
 
   barrier_data = gasneti_malloc_aligned(GASNETI_CACHE_LINE_BYTES, sizeof(gasnete_coll_ibdbarrier_t));
   gasneti_leak_aligned(barrier_data);
@@ -1085,54 +1161,31 @@ static void gasnete_ibdbarrier_init(gasnete_coll_team_t team) {
 
   gasnete_ibdbarrier_lock_init(&barrier_data->barrier_lock);
 
-  /* determine barrier size (number of steps) */
-  for (steps=0, j=1; j < total_ranks; ++steps, j*=2) ;
-
-  barrier_data->barrier_size = steps;
-  barrier_data->barrier_goal = steps << 1;
+  steps = peers->num;
+  barrier_data->barrier_goal = (1+steps) << 1;
 
   if (steps) {
-#if GASNETI_PSHM_BARRIER_HIER
-    gasnet_node_t *nodes = supernode_reps ? supernode_reps : gasneti_pshm_firsts;
-#endif
     int step;
 
     gasneti_assert(gasnete_rdmabarrier_auxseg);
     barrier_data->barrier_inbox = gasnete_rdmabarrier_auxseg[gasneti_mynode].addr;
     gasneti_assert(GASNETE_IBDBARRIER_INBOX_WORDS > 1);
 
-    barrier_data->barrier_peers = gasneti_malloc(steps * sizeof(* barrier_data->barrier_peers));
+    barrier_data->barrier_peers = gasneti_malloc((1+steps) * sizeof(* barrier_data->barrier_peers));
     gasneti_leak(barrier_data->barrier_peers);
   
     for (step = 0; step < steps; ++step) {
-      gasnet_node_t distance, tmp, peer, node;
-
-      distance = (1 << step);
-      tmp = total_ranks - myrank;
-      peer = (distance < tmp) ? (distance + myrank) : (distance - tmp); /* mod N w/o overflow */
-      gasneti_assert(peer < total_ranks);
-
-#if GASNETI_PSHM_BARRIER_HIER
-      if (pshm_bdata) {
-        node = nodes[peer];
-      } else
-#endif
-      {
-        node = GASNETE_COLL_REL2ACT(team, peer);
-      }
-
-      barrier_data->barrier_peers[step].node = node;
-      barrier_data->barrier_peers[step].addr = gasnete_rdmabarrier_auxseg[node].addr;
-    }
+      gasnet_node_t node = peers->fwd[step];
+      barrier_data->barrier_peers[1+step].node = node;
+      barrier_data->barrier_peers[1+step].addr = gasnete_rdmabarrier_auxseg[node].addr;
+    } 
   } else {
-    barrier_data->barrier_slot = barrier_data->barrier_goal;
+    barrier_data->barrier_state = barrier_data->barrier_goal;
   }
 
   gasneti_free(gasnete_rdmabarrier_auxseg);
 
 #if GASNETI_PSHM_BARRIER_HIER
-  gasneti_free(supernode_reps);
-
   if (pshm_bdata && (pshm_bdata->shared->size == 1)) {
     /* With singleton proc on local supernode we can short-cut the PSHM code.
      * This does not require alteration of the barrier_peers[] contructed above
