@@ -73,6 +73,7 @@ static MPI_Comm AMMPI_SPMDMPIComm;
  *  misc helpers
  * ------------------------------------------------------------------------------------ */
 static void flushStreams(const char *context) {
+  static int do_sync = -1;
   if (!context) context = "flushStreams()";
 
   if (fflush(NULL)) { /* passing NULL to fflush causes it to flush all open FILE streams */
@@ -89,11 +90,22 @@ static void flushStreams(const char *context) {
   }
   fsync(STDOUT_FILENO); /* ignore errors for output is a console */
   fsync(STDERR_FILENO); /* ignore errors for output is a console */
+
+  if (do_sync < 0) {
+    /* Approximate match to GASNet's acceptance of 'Y|YES|y|yes|1' */
+    char c, *envval;
+    envval = getenv("GASNET_FS_SYNC");
+    if (NULL == envval) envval = getenv("AMMPI_FS_SYNC");
+    c = envval ? envval[0] : '0';
+    do_sync = ((c == '1') || (c == 'y') || (c == 'Y'));
+  }
+  if (do_sync) {
   #if PLATFORM_OS_MTA
     mta_sync();
   #elif !PLATFORM_OS_CATAMOUNT
     sync();
   #endif
+  }
   ammpi_sched_yield();
 }
 /* ------------------------------------------------------------------------------------ */
@@ -174,12 +186,14 @@ extern int AMMPI_SPMDStartup(int *argc, char ***argv,
   tag_t mytag;
   
   if (AMMPI_SPMDStartupCalled) AMMPI_RETURN_ERR(RESOURCE);
-  if (!argc || !argv) AMMPI_RETURN_ERR(BAD_ARG);
 
   { /* initialize MPI, if necessary */
     int initialized = 0;
     MPI_SAFE(MPI_Initialized(&initialized));
     if (!initialized) {
+  #if MPI_VERSION < 2
+      if (!argc || !argv) AMMPI_RETURN_ERR(BAD_ARG);
+  #endif
       MPI_SAFE(MPI_Init(argc, argv));
     }
   }
